/***
*files.h - definitions/declarations for library functions.
*
*Copyright (c) 1991-1994, Gregg Jennings.  All wrongs reserved.
*   P O Box 200, Falmouth, MA 02541-0200
*
*Purpose:
*   [ANSI]
*
*******************************************************************************/

#ifndef GENERAL_H
#include "general.h"
#endif

#ifndef _PUNCT          /* use MSC equates */
#define _UPPER    0x1   /* upper case letter */
#define _LOWER    0x2   /* lower case letter */
#define _DIGIT    0x4   /* digit[0-9] */
#define _SPACE    0x8   /* tab, carriage return, newline, */
                        /* vertical tab or form feed */
#define _PUNCT    0x10  /* punctuation character */
#define _CONTROL  0x20  /* control character */
#define _BLANK    0x40  /* space char */
#define _HEX      0x80  /* hexadecimal digit */
#endif

extern unsigned char hexbyte(int a,int b);
extern void print(const char *s);
extern void printc(const char *s);
extern void bprint(const char *s);
extern void put(unsigned int i,int c);
extern int pn(unsigned int n,int base);
extern int pln(unsigned long n, int base);
extern void pnls(unsigned int n,unsigned int length,int base);
extern void pnlz(unsigned int n,unsigned int length,int base);
extern void plnls(unsigned long n,unsigned int length,int base);
extern void plnlz(unsigned long n,unsigned int length,int base);

extern int getlen(unsigned int n,int base);
extern int getlength(unsigned long i,int base);
extern int getnum(unsigned int max,unsigned int *num,int base);
extern int getnumber(unsigned long max,void *num,int base);

extern int getstr(char *s,unsigned int len,int mask);
extern int getver(char *s,int mode);
enum GETVER_MODE { NONE, MOV_YN, MOV_ARG, CLR_ARG=5, CLR_YN };
extern int convert(char *a);
extern void dir(void );
extern void pname(char *s);
extern int search(unsigned char *s,unsigned char *t,unsigned int size,unsigned int len);
extern int isearch(unsigned char *s,unsigned char *t,unsigned int size,unsigned int len);
extern int change(unsigned char *buffer,unsigned int bufend,int base);
extern void dispval(int x,int len,int base,int c);
extern int dump(unsigned char *buffer,unsigned int bfptr,unsigned int bufend,unsigned int len,int base);
extern void dumpf(register unsigned char *buffer,int size, int base);
extern int send(int c);
extern void pause(void );
extern void conout(int c);
extern void charout(int c);
extern void conouta(int c, int atr);
extern void cursoron(void);
extern void cursoroff(void);
extern void getcursor(byte *r, byte *c);
extern void setcursor(byte r, byte c);
extern void cursor(int);
#define savecursor() cursor(1)
#define restcursor() cursor(0)
extern void clreol(void);
extern int conin(void);
extern int input(void);
extern int conflag(void);
extern void curright(void);
extern void curleft(void);

/* macros to make displaying of numbers easier to implement */

#ifndef Radix                 /* Radix is a global variable */
#define Radix   kw[16].opt    /* DISKED uses a structure entry */
#endif                        /* This is here for non-DISKED usage */
                              /* define Radix for you own use */

#define getbyte(n)   getnum(255,n,Radix)
#define getint(n)    getnum(32767,n,Radix)
#define getuint(n)   getnum(65535,n,Radix)
#define getlong(n)   getnumber(2147483647L,n,Radix)
#define getnumb(m,n) getnum(m,n,Radix)
#define getlnumb(m,n) getnumber((unsigned long)(m),n,Radix)
#define len(n)       getlen(n,Radix)
#define length(n)    getlength((unsigned long)(n),Radix)

#define putbyte(n)   pn((int)(n),Radix)
#define puthex(n)    pnlz(n,4,16)
#define putint(n)    pn(n,Radix)
#define putuint(n)   pln((unsigned long)(n),Radix)
#define putlong(n)   pln((unsigned long)(n),Radix)
#define zputbyte(n)  pnlz((unsigned int)(n),3,Radix)
